==> pan-0.90.tar.bz2: pan-0.90/pan/tasks/queue.cc <==
/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <glib.h>
#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include <pan/general/messages.h>
#include <pan/data/server-info.h>
#include "queue.h"
#include "task.h"

/***
****
***/

using namespace pan;

Queue :: Queue (const ServerInfo   & server_info,
                TaskArchive        & archive,
                Socket::Creator    * socket_creator,
                int                  msec):
  _server_info (server_info),
  _is_online (true),
  _socket_creator (socket_creator),
  _archive (archive)
{
  task_list_t tasks;
  _archive.load_tasks (tasks);
  add_tasks (tasks);
}

Queue :: ~Queue ()
{
  foreach (pools_t, _pools, it)
    delete it->second;
  _pools.clear ();

  foreach (items_t, _items, it)
    delete *it;
}

void
Queue :: on_nntp_available (const Quark& server)
{
  Task * task = find_first_task_needing_server (server);
  if (task != 0)
    process_task (task);
}

NNTP_Pool&
Queue :: get_pool (const Quark& servername)
{
  NNTP_Pool * pool (0);

  pools_t::iterator it = _pools.find (servername);
  if (it != _pools.end())
  {
    pool = it->second;
  }
  else // have to build one
  {
    std::string username, password;
    _server_info.get_server_auth (servername, username, password);
    pool = new NNTP_Pool (servername, _socket_creator, username, password);
    
    std::string host;
    int port;
    const bool got_address (_server_info.get_server_addr (servername, host, port));
    if (got_address)
      pool->set_address (host, port);

    pool->set_max_connections (_server_info.get_server_limits (servername));

    pool->add_listener (this);
    _pools[servername] = pool;
  }

  return *pool;
}

/***
****
***/

void
Queue :: upkeep ()
{
  debug ("queue " << this << " upkeep...");

  // process the tasks...
  items_t tmp (_items);
  foreach (items_t, tmp, it)  
    process_task (*it);

  // if our number of connections has changed, fire an event...
  static int previous_count (-1);
  int count (0);
  foreach (pools_t, _pools, it) {
    int active, idle, pending, max;
    it->second->get_counts (active, idle, pending, max);
    count += (active + idle + pending);
  }
  const bool count_changed (previous_count != count);
  previous_count = count;
  if (count_changed)
    fire_connection_count_changed (count);

  // if our number of tasks has changed, fire an event...
  static int prev_active (-1);
  static int prev_total (-1);
  const int total (size());
  std::set<Task*> active_tasks;
  foreach_const (nntp_to_task_t, _nntp_to_task, it)
    active_tasks.insert (it->second);
  const int active (active_tasks.size());
  const bool counts_changed (active!=prev_active || total!=prev_total);
  prev_active = active;
  prev_total = total;
  if (counts_changed)
    fire_size_changed (active, total);
}

void
Queue :: process_task (Task * task)
{
  pan_return_if_fail (task!=0);

  debug ("in process_task with a task of type " << task->get_type());

  const Task::State& state (task->get_state());

  if (state._work == Task::COMPLETED)
  {
    debug ("completed");
    remove_task (task);
  }
  else if (_removing.count(task))
  {
    debug ("removing");
    remove_task (task);
  }
  else if (_stopped.count(task))
  {
    debug ("stopped");
    // do nothing
  }
  else if (state._health == FAIL)
  {
    debug ("fail");
    // do nothing
  }
  else if (state._work == Task::WORKING)
  {
    debug ("working");
    // do nothing
  }
  else // _work == NEED_NNTP
  {
    debug ("need nntp");
    const quarks_t servers (state._servers);
    foreach (quarks_t, servers, it)
        get_pool(*it).request_nntp ();
    Quark server;
    if (find_best_server (servers, server))
    {
      NNTP_Pool& pool (get_pool (server));
      NNTP * nntp (pool.check_out ());
      if (nntp) {
        const bool was_active (task_is_active (task));
        _nntp_to_task[nntp] = task; // it's active now...
        if (!was_active)
          fire_task_active_changed (task, true);
        nntp->_socket->reset_speed_counter ();
        task->give_nntp (this, nntp);
      }
    }
  }
}

/***
****
***/

Task*
Queue :: find_first_task_needing_server (const Quark& server)
{
  Task * retval (0);

  foreach (items_t, _items, it) {
    const Task::State state ((*it)->get_state ());
    if ((state._health != FAIL) && (state._work == Task::NEED_NNTP)) {
      retval = *it;
      break;
    }
  }

  return retval;
}

bool
Queue :: find_best_server (const quarks_t& servers, Quark& setme)
{
  int max_score (0);
  Quark best_server;

  foreach (quarks_t, servers, it)
  {
    const Quark& server (*it);
    const NNTP_Pool& pool (get_pool(server));

    int score;
    if (!_is_online || !pool.is_online()) { // if this, or all, servers are offline...
      score = 0;
    } else {
      int active, idle, pending, max;
      pool.get_counts (active, idle, pending, max);
      const int empty_slots (max - (idle+active));
      score = idle*10 + empty_slots;
    }

    if (score > max_score) {
      max_score = score;
      best_server = server;
    }
  }

  if (max_score)
    setme = best_server;

  return max_score!=0;
}

/***
****
***/

void
Queue :: fire_items_added (task_list_t& tasks, int index)
{
  _archive.save_tasks (_items);

  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_tasks_added (*this, tasks, index); 
}
void
Queue :: fire_item_removed (Task*& task, int index)
{
  // save a list of the tasks w/o 'task'
  task_list_t tmp (_items);
  tmp.erase (tmp.begin() + index);
  _archive.save_tasks (tmp);

  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_task_removed (*this, *task, index);
}
void
Queue :: fire_item_moved (Task*& task, int new_index, int old_index)
{
  std::cerr << "task " << (void*)task << " moved from "<< old_index << " to " << new_index << std::endl;
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_task_moved (*this, *task, new_index, old_index);
}

void
Queue :: fire_connection_count_changed (int count)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_connection_count_changed (*this, count);
}
void
Queue :: fire_size_changed (int active, int total)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_size_changed (*this, active, total);
}
void
Queue :: fire_online_changed (bool online)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_online_changed (*this, online);
}
void
Queue :: fire_task_active_changed (Task * task, bool active)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_task_active_changed (*this, *task, active);
}
void
Queue :: add_listener (Listener * l)
{
   _listeners.insert (l);
}
void
Queue :: remove_listener (Listener * l)
{
   _listeners.erase (l);
}

/***
****
***/

void
Queue :: set_online (bool online)
{
  _is_online = online;
  fire_online_changed (_is_online);
}

void
Queue :: stop_tasks (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    if (index_of (task) != -1)
      _stopped.insert (task);
  }
}

void
Queue :: restart_tasks (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    if (index_of(task) != -1) {
      _stopped.erase (task);
      process_task (task);
    }
  }
}

void
Queue :: add_task (Task * task)
{
  add (task);
  //const int index (add (task));
  //fire_task_added (task, index);
  upkeep ();
}

void
Queue :: add_tasks (const task_list_t& tasks)
{
  add (tasks);
  // FIXME fire
  upkeep ();
}

void
Queue :: remove_latest_task ()
{
  if (!empty())
    remove_task (_items[size()-1]);
}

bool
Queue :: task_is_active (const Task * task) const
{
  bool task_has_nntp (false);
  foreach_const (nntp_to_task_t, _nntp_to_task, it)
    if ((task_has_nntp = task==it->second))
      break;
  return task_has_nntp;
}

void
Queue :: remove_tasks (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it)
    remove_task (*it);
}

void
Queue :: remove_task (Task * task)
{
  const int index (index_of (task));
  pan_return_if_fail (index != -1);

  if (task_is_active (task)) // wait for the NNTPs to finish
  {
    debug ("can't delete this task right now because it's got server connections");
    _removing.insert (task);
  }
  else // no NNTPs working, we can remove right now.
  {
    _stopped.erase (task);
    _removing.erase (task);
    remove (index);
    //fire_task_removed (task, index);
    delete task;
  }
}

void
Queue :: get_tasks (tasks_t& tasks)
{
  // get the state of each task
  typedef std::map<Task*,TaskState> task_to_state_t;
  task_to_state_t states;
  foreach (items_t, _items, it) states[*it] = QUEUED;
  foreach (std::set<Task*>, _removing, it) states[*it] = REMOVING;
  foreach (std::set<Task*>, _stopped, it) states[*it] = STOPPED;
  foreach (nntp_to_task_t, _nntp_to_task, it) states[it->second] = RUNNING;

  tasks.tasks = _items;
  tasks.states = states;
}

void
Queue :: check_in (NNTP * nntp, bool is_ok)
{
  Task * task (_nntp_to_task[nntp]);

  // if the same task still needs a connection to this server,
  // shoot it straight back.  This can be a lot faster than
  // returning the NNTP to the pool and checking it out again.
  const Task::State state (task->get_state ());
  if (is_ok
    && (state._health!=FAIL)
    && (state._work==Task::NEED_NNTP)
    && !_removing.count(task)
    && state._servers.count(nntp->_server)
    && find_first_task_needing_server(nntp->_server)==task)
  {
    task->give_nntp (this, nntp);
  }
  else
  {
    // take care of our nntp counting
    Task * task (_nntp_to_task[nntp]);
    _nntp_to_task.erase (nntp);

    // take care of the task's state
    const bool is_active (task_is_active (task));
    if (!is_active) // if it's not active anymore...
      fire_task_active_changed (task, is_active);

    // return the nntp to the pool
    const Quark& servername (nntp->_server);
    NNTP_Pool& pool (get_pool (servername));
    pool.check_in (nntp, is_ok);
  }
}

void
Queue :: move_up (const task_list_t& tasks)
{
  for (task_list_t::const_reverse_iterator it(tasks.rbegin()), end(tasks.rend()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_up (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

void
Queue :: move_down (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_down (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

void
Queue :: move_top (const task_list_t& tasks)
{
  for (task_list_t::const_reverse_iterator it(tasks.rbegin()), end(tasks.rend()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_top (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

void
Queue :: move_bottom (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_bottom (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

double
Queue :: get_speed_KiBps () const
{
  double KiBps (0.0);
  foreach_const (nntp_to_task_t, _nntp_to_task, it)
    KiBps += it->first->_socket->get_speed_KiBps ();
  return KiBps;
}

void
Queue :: get_task_speed_KiBps (const Task * task, double& setme_KiBps, int& setme_connections) const
{
  double KiBps (0.0);
  int connections (0);
  foreach_const (nntp_to_task_t, _nntp_to_task, it) {
    if (it->second==task) {
      ++connections;
      KiBps += it->first->_socket->get_speed_KiBps ();
    }
  }

  setme_KiBps = KiBps;
  setme_connections = connections;
}

==> pan-0.92.tar.bz2: pan-0.92/pan/tasks/queue.cc <==
/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/*
 * Pan - A Newsreader for Gtk+
 * Copyright (C) 2002-2006  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>
#include <glib.h>
#include <pan/general/debug.h>
#include <pan/general/foreach.h>
#include <pan/general/messages.h>
#include <pan/data/server-info.h>
#include "queue.h"
#include "task.h"

/***
****
***/

using namespace pan;

Queue :: Queue (const ServerInfo   & server_info,
                TaskArchive        & archive,
                Socket::Creator    * socket_creator,
                int                  msec):
  _server_info (server_info),
  _is_online (true),
  _socket_creator (socket_creator),
  _archive (archive)
{
  task_list_t tasks;
  _archive.load_tasks (tasks);
  add_tasks (tasks);
}

Queue :: ~Queue ()
{
  foreach (pools_t, _pools, it)
    delete it->second;
  _pools.clear ();

  foreach (items_t, _items, it)
    delete *it;
}

void
Queue :: on_nntp_available (const Quark& server)
{
  Task * task = find_first_task_needing_server (server);
  if (task != 0)
    process_task (task);
}

NNTP_Pool&
Queue :: get_pool (const Quark& servername)
{
  NNTP_Pool * pool (0);

  pools_t::iterator it = _pools.find (servername);
  if (it != _pools.end())
  {
    pool = it->second;
  }
  else // have to build one
  {
    std::string username, password;
    _server_info.get_server_auth (servername, username, password);
    pool = new NNTP_Pool (servername, _socket_creator, username, password);
    
    std::string host;
    int port;
    const bool got_address (_server_info.get_server_addr (servername, host, port));
    if (got_address)
      pool->set_address (host, port);

    pool->set_max_connections (_server_info.get_server_limits (servername));

    pool->add_listener (this);
    _pools[servername] = pool;
  }

  return *pool;
}

/***
****
***/

void
Queue :: upkeep ()
{
  debug ("queue " << this << " upkeep...");

  // process the tasks...
  items_t tmp (_items);
  foreach (items_t, tmp, it)  
    process_task (*it);

  // if our number of connections has changed, fire an event...
  static int previous_count (-1);
  int count (0);
  foreach (pools_t, _pools, it) {
    int active, idle, pending, max;
    it->second->get_counts (active, idle, pending, max);
    count += (active + idle + pending);
  }
  const bool count_changed (previous_count != count);
  previous_count = count;
  if (count_changed)
    fire_connection_count_changed (count);

  // if our number of tasks has changed, fire an event...
  static int prev_active (-1);
  static int prev_total (-1);
  const int total (size());
  std::set<Task*> active_tasks;
  foreach_const (nntp_to_task_t, _nntp_to_task, it)
    active_tasks.insert (it->second);
  const int active (active_tasks.size());
  const bool counts_changed (active!=prev_active || total!=prev_total);
  prev_active = active;
  prev_total = total;
  if (counts_changed)
    fire_size_changed (active, total);
}

void
Queue :: process_task (Task * task)
{
  pan_return_if_fail (task!=0);

  debug ("in process_task with a task of type " << task->get_type());

  const Task::State& state (task->get_state());

  if (state._work == Task::COMPLETED)
  {
    debug ("completed");
    remove_task (task);
  }
  else if (_removing.count(task))
  {
    debug ("removing");
    remove_task (task);
  }
  else if (_stopped.count(task))
  {
    debug ("stopped");
    // do nothing
  }
  else if (state._health == FAIL)
  {
    debug ("fail");
    // do nothing
  }
  else if (state._work == Task::WORKING)
  {
    debug ("working");
    // do nothing
  }
  else // _work == NEED_NNTP
  {
    debug ("need nntp");
    const quarks_t servers (state._servers);
    foreach (quarks_t, servers, it)
        get_pool(*it).request_nntp ();
    Quark server;
    if (find_best_server (servers, server))
    {
      NNTP_Pool& pool (get_pool (server));
      NNTP * nntp (pool.check_out ());
      if (nntp) {
        const bool was_active (task_is_active (task));
        _nntp_to_task[nntp] = task; // it's active now...
        if (!was_active)
          fire_task_active_changed (task, true);
        nntp->_socket->reset_speed_counter ();
        task->give_nntp (this, nntp);
      }
    }
  }
}

/***
****
***/

Task*
Queue :: find_first_task_needing_server (const Quark& server)
{
  Task * retval (0);

  foreach (items_t, _items, it) {
    const Task::State state ((*it)->get_state ());
    if ((state._health != FAIL) && (state._work == Task::NEED_NNTP)) {
      retval = *it;
      break;
    }
  }

  return retval;
}

bool
Queue :: find_best_server (const quarks_t& servers, Quark& setme)
{
  int max_score (0);
  Quark best_server;

  foreach (quarks_t, servers, it)
  {
    const Quark& server (*it);
    const NNTP_Pool& pool (get_pool(server));

    int score;
    if (!_is_online || !pool.is_online()) { // if this, or all, servers are offline...
      score = 0;
    } else {
      int active, idle, pending, max;
      pool.get_counts (active, idle, pending, max);
      const int empty_slots (max - (idle+active));
      score = idle*10 + empty_slots;
    }

    if (score > max_score) {
      max_score = score;
      best_server = server;
    }
  }

  if (max_score)
    setme = best_server;

  return max_score!=0;
}

/***
****
***/

void
Queue :: fire_items_added (task_list_t& tasks, int index)
{
  _archive.save_tasks (_items);

  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_tasks_added (*this, tasks, index); 
}
void
Queue :: fire_item_removed (Task*& task, int index)
{
  // save a list of the tasks w/o 'task'
  task_list_t tmp (_items);
  tmp.erase (tmp.begin() + index);
  _archive.save_tasks (tmp);

  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_task_removed (*this, *task, index);
}
void
Queue :: fire_item_moved (Task*& task, int new_index, int old_index)
{
  std::cerr << "task " << (void*)task << " moved from "<< old_index << " to " << new_index << std::endl;
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_task_moved (*this, *task, new_index, old_index);
}

void
Queue :: fire_connection_count_changed (int count)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_connection_count_changed (*this, count);
}
void
Queue :: fire_size_changed (int active, int total)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_size_changed (*this, active, total);
}
void
Queue :: fire_online_changed (bool online)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_online_changed (*this, online);
}
void
Queue :: fire_task_active_changed (Task * task, bool active)
{
  for (lit it(_listeners.begin()), end(_listeners.end()); it!=end; )
    (*it++)->on_queue_task_active_changed (*this, *task, active);
}
void
Queue :: add_listener (Listener * l)
{
   _listeners.insert (l);
}
void
Queue :: remove_listener (Listener * l)
{
   _listeners.erase (l);
}

/***
****
***/

void
Queue :: set_online (bool online)
{
  _is_online = online;
  fire_online_changed (_is_online);
}

void
Queue :: stop_tasks (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    if (index_of (task) != -1)
      _stopped.insert (task);
  }
}

void
Queue :: restart_tasks (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    if (index_of(task) != -1) {
      _stopped.erase (task);
      process_task (task);
    }
  }
}

void
Queue :: add_task (Task * task)
{
  add (task);
  //const int index (add (task));
  //fire_task_added (task, index);
  upkeep ();
}

void
Queue :: add_tasks (const task_list_t& tasks)
{
  add (tasks);
  // FIXME fire
  upkeep ();
}

void
Queue :: remove_latest_task ()
{
  if (!empty())
    remove_task (_items[size()-1]);
}

bool
Queue :: task_is_active (const Task * task) const
{
  bool task_has_nntp (false);
  foreach_const (nntp_to_task_t, _nntp_to_task, it)
    if ((task_has_nntp = task==it->second))
      break;
  return task_has_nntp;
}

void
Queue :: remove_tasks (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it)
    remove_task (*it);
}

void
Queue :: remove_task (Task * task)
{
  const int index (index_of (task));
  pan_return_if_fail (index != -1);

  if (task_is_active (task)) // wait for the NNTPs to finish
  {
    debug ("can't delete this task right now because it's got server connections");
    _removing.insert (task);
  }
  else // no NNTPs working, we can remove right now.
  {
    _stopped.erase (task);
    _removing.erase (task);
    remove (index);
    //fire_task_removed (task, index);
    delete task;
  }
}

void
Queue :: get_tasks (tasks_t& tasks)
{
  // get the state of each task
  typedef std::map<Task*,TaskState> task_to_state_t;
  task_to_state_t states;
  foreach (items_t, _items, it) states[*it] = QUEUED;
  foreach (std::set<Task*>, _removing, it) states[*it] = REMOVING;
  foreach (std::set<Task*>, _stopped, it) states[*it] = STOPPED;
  foreach (nntp_to_task_t, _nntp_to_task, it) states[it->second] = RUNNING;

  tasks.tasks = _items;
  tasks.states = states;
}

void
Queue :: check_in (NNTP * nntp, bool is_ok)
{
  Task * task (_nntp_to_task[nntp]);

  // if the same task still needs a connection to this server,
  // shoot it straight back.  This can be a lot faster than
  // returning the NNTP to the pool and checking it out again.
  const Task::State state (task->get_state ());
#if 0
  std::cerr << LINE_ID << " task " << task << " returned a server " << nntp->_server
    << " is_ok: " << is_ok
    << " state._health!=FAIL " << (state._health!=FAIL)
    << " state._work==Task::NEED_NNTP " << (state._work==Task::NEED_NNTP)
    << " !_removing.count(task) " << !_removing.count(task)
    //<< " state._servers.count(nntp->_server) " << state._servers.count(nntp->_server) << std::endl;
    //<< " find_first_task_needing_server(nntp->_server)==task) " << (find_first_task_needing_server(nntp->_server)==task) << std::endl;
#endif
         
  if (is_ok
    && (state._health!=FAIL)
    && (state._work==Task::NEED_NNTP)
    && !_removing.count(task)
    && state._servers.count(nntp->_server))
    //&& find_first_task_needing_server(nntp->_server)==task)
  {
    task->give_nntp (this, nntp);
  }
  else
  {
    // take care of our nntp counting
    Task * task (_nntp_to_task[nntp]);
    _nntp_to_task.erase (nntp);

    // take care of the task's state
    const bool is_active (task_is_active (task));
    if (!is_active) // if it's not active anymore...
      fire_task_active_changed (task, is_active);

    // return the nntp to the pool
    const Quark& servername (nntp->_server);
    NNTP_Pool& pool (get_pool (servername));
    pool.check_in (nntp, is_ok);
  }
}

void
Queue :: move_up (const task_list_t& tasks)
{
  for (task_list_t::const_reverse_iterator it(tasks.rbegin()), end(tasks.rend()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_up (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

void
Queue :: move_down (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_down (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

void
Queue :: move_top (const task_list_t& tasks)
{
  for (task_list_t::const_reverse_iterator it(tasks.rbegin()), end(tasks.rend()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_top (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

void
Queue :: move_bottom (const task_list_t& tasks)
{
  for (task_list_t::const_iterator it(tasks.begin()), end(tasks.end()); it!=end; ++it) {
    Task * task (*it);
    const int old_pos (index_of (task));
    TaskSet::move_bottom (old_pos);
    //const int new_pos (index_of (task));
    //if (old_pos != new_pos)
     // fire_task_moved (task, new_pos, old_pos);
  }
}

double
Queue :: get_speed_KiBps () const
{
  double KiBps (0.0);
  foreach_const (nntp_to_task_t, _nntp_to_task, it)
    KiBps += it->first->_socket->get_speed_KiBps ();
  return KiBps;
}

void
Queue :: get_task_speed_KiBps (const Task * task, double& setme_KiBps, int& setme_connections) const
{
  double KiBps (0.0);
  int connections (0);
  foreach_const (nntp_to_task_t, _nntp_to_task, it) {
    if (it->second==task) {
      ++connections;
      KiBps += it->first->_socket->get_speed_KiBps ();
    }
  }

  setme_KiBps = KiBps;
  setme_connections = connections;
}
